
#include <stdio.h>
#include <stdlib.h>
#include <iostream>

using namespace std;

#include <math.h>
#include <float.h>

#include "Graph.h"
#include "RandomField.h"

CAnnealing::CAnnealing(int nV, int K, double psi) {

  mFalsePositive = 0.0;
  mFalseNegative = 0.0;
  mPsi = psi;
  mNumVars = nV;
  mNumClusters = K;
  
  mQ = new double*[nV];

  for(int i=0; i < nV; i++) {
    mQ[i] = new double[K];
  }
  
  for(int i=0; i < nV; i++) {
    for(int j=0; j < K; j++) {
      mQ[i][j] = 0.0;
    }
  }
}

CAnnealing::~CAnnealing(){

  for(int i=0; i < mNumVars; i++) {
    delete [] mQ[i];
  }
  
  delete [] mQ;
}

#include <math.h>
void CAnnealing::initialize() {

  double r;

  for(int i=0; i < mNumVars; i++) {
    double s=0.0;
    for(int j=0; j < mNumClusters; j++) {
      r = 1.0 + (double)rand()/(1.0 + (double)RAND_MAX);
      mQ[i][j] = r;
      s += r;
    }

    for(int j=0; j < mNumClusters; j++) {
      mQ[i][j] /= s;
    }
  }
}

double CAnnealing::reestimate(CGraph &G, double gamma) {

  double err=0.0;
  double *Cost = new double[mNumClusters];
  double *Q = new double[mNumClusters];
  
  for(int i=0; i < mNumVars; i++) {
    
    double minCost = DBL_MAX;

    for(int g=0; g < mNumClusters; g++) {      

      Cost[g] = 0.0;

      for(int j=0; j < G.mNodes[i].mAdjacency.size(); j++) {	
	double t, s;
	int v = G.mNodes[i].mAdjacency[j].v;
	t = (double) G.mNodes[i].mAdjacency[j].t;
	s = (double) G.mNodes[i].mAdjacency[j].s;

	Cost[g] += mQ[v][g]*(t-s) + (1.0-mQ[v][g])*mPsi*s; 
      }

      if (minCost > Cost[g]) {
	minCost = Cost[g];
      }
    }

    double Z = 0.0;
    for(int g=0; g < mNumClusters; g++) {
      Cost[g] -= minCost;
      Q[g] = exp(-gamma*Cost[g]); 
      Z += Q[g];
    }
    
    double Z2 = 0.0;
    for(int g=0; g < mNumClusters; g++) {      
      double r = (double)rand()/(1.0 + (double)RAND_MAX);
      Q[g] = (Q[g]/Z) + r*1e-5;
      
      // assert(!__isnand(Q[g]));      
      Z2 += Q[g];      
    }

    for(int g=0; g < mNumClusters; g++) {      
      double q = Q[g]/Z2;
      err += fabs(mQ[i][g] - q);
      mQ[i][g] = q;
    }
  }

  delete [] Q;
  delete [] Cost;
  return err;
}

double CAnnealing::coOccurrence(CGraph &G) {

  double L = 0.0;  
  for(int i=0; i < mNumVars; i++) {
    for(int j=0; j < G.mNodes[i].mAdjacency.size(); j++) {

      int v = G.mNodes[i].mAdjacency[j].v;

      double X = 0.0;
      for(int g=0; g < mNumClusters; g++) {
	X += mQ[i][g] * mQ[v][g];
      }

      double t = (double) G.mNodes[i].mAdjacency[j].t; 
      double s = (double) G.mNodes[i].mAdjacency[j].s; 

      t = X*(t-s) + (1.0-X)*mPsi*s;
      L += t;
    }
  }
  return L;
}

double CAnnealing::Likelihood(CGraph &G) {

  double alpha, beta;
  alpha = -log(mFalseNegative) + log(1-mFalsePositive);
  beta = -log(mFalsePositive) + log(1-mFalseNegative);

  double L = 0.0;  

  for(int i=0; i < G.mNodes.size(); i++) {
    for(int j=0; j < G.mNodes[i].mAdjacency.size(); j++) {

      int v = G.mNodes[i].mAdjacency[j].v;

      double Q;
      if (mGroups[i] == mGroups[v]) {
	Q = 1.0;
      } else {
	Q = 0.0;
      }

      double t = (double) G.mNodes[i].mAdjacency[j].t; 
      double s = (double) G.mNodes[i].mAdjacency[j].s; 

      double A = s*(Q*(-log(1-mFalseNegative)) + (1.0-Q)*(-log(mFalsePositive)));
      double B = (t-s)*(Q*(-log(mFalseNegative)) + (1.0-Q)*(-log(1-mFalsePositive)));

      // double lambda = s*beta*(1-Q) - s*log(1-fn) + (t-s)*alpha*Q - (t-s)*log(1-fp);
      L += (A + B);
    }
  }
  return L;
}

double CAnnealing::UpdatePsi() {

  double C;

  if (fabs(mFalsePositive + mFalseNegative - 1.0) > 1e-6) {
    C = (-log(mFalsePositive) + log(1-mFalseNegative))/(-log(mFalseNegative) + log(1-mFalsePositive));
    
    cout << "FN = " << mFalseNegative << endl;
    cout << "FP = " << mFalsePositive << endl;
    cout << "C = " << C << endl;

  } else {

    C = mFalseNegative/(1.0 - mFalseNegative);

    cout << "FN = " << mFalseNegative << endl;
    cout << "FP = " << mFalsePositive << endl;
    cout << "Approx. C. = " << C << endl;

    assert(!__isnan(mPsi));

  }
  return C;
}

double CAnnealing::RateEstimation(CGraph &G) {

  double A, B, C, D;

  A = 0.0;
  B = 0.0;
  C = 0.0;
  D = 0.0;

  for(int i=0; i < mNumVars; i++) {
    for(int j=0; j < G.mNodes[i].mAdjacency.size(); j++) {

      int v = G.mNodes[i].mAdjacency[j].v;

      double Q = 0.0;
      for(int g=0; g < mNumClusters; g++) {
	Q += mQ[i][g] * mQ[v][g];
      }

      double t = (double) G.mNodes[i].mAdjacency[j].t; 
      double s = (double) G.mNodes[i].mAdjacency[j].s; 
      
      A += Q*(t-s);
      B += Q*(t);
      
      C += (1.0 - Q)*s;
      D += (1.0 - Q)*t;      
    }
  }
  
  double err;
  
  mFalseNegative = A/B;
  mFalsePositive = C/D;

  double psi = UpdatePsi();
  err = fabs(mPsi - psi);
  mPsi = psi; 

  return err;
}

double CAnnealing::assignment(CGraph &G, int maxIteration) {

  double L;

  initialize();
  L = coOccurrence(G);  
  cout << "Initial likelihood = " << L << endl;


  int i=0;
  double err;

  double gamma = 0.01;
  do {
    for(i=0; i < 10; i++) {
      err = reestimate(G, gamma);
      cout << "Gamma = " << gamma << " iteration " << i << ", error = " << err << endl;
    }
  

    do {
      err = reestimate(G, gamma);
      cout << "Gamma = " << gamma << " iteration " << i << ", error = " << err << endl;
      i++;
    } while( err > 0.01 && i < 100);
    gamma *= 2.1;    

  } while( gamma < 1000.0 );
  
  RateEstimation(G);    
  
  mGroups.resize(mNumVars);
  
  for(int i = 0; i < mNumVars; i++) {
    double maxQ = 0.0;
    mGroups[i] = -1;
    for(int g = 0; g < mNumClusters; g++) {
      if (mQ[i][g] > maxQ) {
	maxQ = mQ[i][g];
	mGroups[i] = g+1;
      }
    }
  }

  L = Likelihood(G);
  mLambda = L;
  cout << "Likelihood = " << L << endl;

  return L;
}
